import string


# A way of describing arithmetic:

# number = Many(digits)
# factor = ("(" + expr + ")") | number
# term = factor + Many("*/" + factor)
# expr = term + Many("+-" + term)


def oper(left, rest):
    # externalize the operations to a function
    for op, right in rest:
        if op == "*":
            left *= right
        elif op == "/":
            left /= right
        elif op == "+":
            left += right
        else:
            left -= right
    return left


def number(pos, data):
    # number = Many(digits).map(to_int)
    rest = []

    # first build up a list
    while data[pos] and data[pos] in string.digits:
        rest.append(data[pos])
        pos += 1

    # then delegate to int to perform the operation
    return pos, int("".join(rest))


def factor(pos, data):
    # factor = ("(" + expr + ")") | number
    if data[pos] == "(":
        pos, res = expr(pos + 1, data)
        assert data[pos] == ")", "Unmatched paren."
        return pos + 1, res
    return number(pos, data)


def term(pos, data):
    # term = factor + Many("*/" + factor)
    pos, left = factor(pos, data)

    # first build up a list
    rest = []
    while data[pos] and data[pos] in "*/":
        op = data[pos]
        pos, right = factor(pos + 1, data)
        rest.append([op, right])

    # then delegate to oper to perform the operations
    return pos, oper(left, rest)


def expr(pos, data):
    # expr = term + Many("+-" + term)
    pos, left = term(pos, data)

    # first build up a list
    rest = []
    while data[pos] and data[pos] in "+-":
        op = data[pos]
        pos, right = term(pos + 1, data)
        rest.append([op, right])

    # then delegate to oper to perform the operations
    return pos, oper(left, rest)


def evaluate(data):
    data = list(data)
    data.append(None)
    return expr(0, data)[1]



"""
This module allows parsers to construct data with a common representation that
is compatible with parsr.query.dsl.

The model allows duplicate keys, and it allows values with *unnamed* attributes
and recursive substructure. This is a common model for many kinds of
configuration.

Simple key/value pairs can be represented as a key with a value that has a
single attribute. Most dictionary shapes used to represent configuration are
made of keys with simple values (key/single attr), lists of simple values
(key/multiple attrs), or nested dictionaries (key/substructure).

Something like XML allows duplicate keys, and it allows values to have named
attributes and substructure. This module doesn't cover that case.

`Entry` and `Result` have overloaded __getitem__ functions that respond to
queries from the parsr.query.dsl module. This allows their instances to be
accessed like simple dictionaries, but the key passed to `[]` is converted to a
query of all child instances instead of a simple lookup.
"""
import operator
from functools import partial
from itertools import chain
from parsr.query.boolean import All, Any, Boolean, lift, lift2, TRUE


class Entry:
    def __init__(self, name=None, attrs=None, children=None, lineno=None):
        self.name = name
        self.attrs = attrs or []
        self.children = children or []
        self.parent = None
        self.lineno = lineno
        for c in self.children:
            c.parent = self

    @property
    def string_value(self):
        t = " ".join(["%s"] * len(self.attrs))
        return t % tuple(self.attrs)

    @property
    def value(self):
        if len(self.attrs) == 1:
            return self.attrs[0]
        return self.attrs or None

    @property
    def root(self):
        p = self
        while p.parent is not None and p.parent.parent is not None:
            p = p.parent
        return p

    @property
    def grandchildren(self):
        return list(chain.from_iterable(c.children for c in self.children))

    def find(self, *queries, roots=False):
        query = compile_queries(*queries)
        return find(query, self.children, roots=roots)

    def __contains__(self, key):
        return len(self[key]) > 0

    def __len__(self):
        return len(self.children)

    def __getitem__(self, query):
        if isinstance(query, (int, slice)):
            return self.children[query]
        query = desugar(query)
        return Result(children=[c for c in self.children if query.test(c)])

    def __repr__(self):
        return f"{self.name}: {self.string_value}"


class Result(Entry):
    def __init__(self, children=None):
        super().__init__()
        self.children = children or []

    @property
    def string_value(self):
        v = self.value
        t = " ".join(["%s"] * len(v))
        return t % tuple(v)

    @property
    def value(self):
        if len(self.children) == 1:
            return self.children[0].value
        raise Exception("More than one value to return.")

    def find(self, *queries, roots=False):
        query = compile_queries(*queries)
        return find(query, self.grandchildren, roots=roots)

    def __getitem__(self, query):
        if isinstance(query, (int, slice)):
            return self.children[query]
        query = desugar(query)
        return Result(children=[c for c in self.grandchildren if query.test(c)])

    def __repr__(self):
        return f"<Result>"


def from_dict(orig):
    def inner(d):
        result = []
        for k, v in d.items():
            if isinstance(v, dict):
                result.append(Entry(name=k, children=inner(v)))
            elif isinstance(v, list):
                res = [from_dict(i) if isinstance(i, dict) else i for i in v]
                if res:
                    if isinstance(res[0], Entry):
                        result.append(Entry(name=k, children=res))
                    else:
                        result.append(Entry(name=k, attrs=res))
                else:
                    result.append(Entry(name=k))
            else:
                result.append(Entry(name=k, attrs=[v]))
        return result
    return Entry(children=inner(orig))


class EntryQuery:
    def __init__(self, expr):
        self.expr = expr

    def test(self, node):
        return self.expr.test(node)


class NameQuery(EntryQuery):
    def test(self, node):
        return self.expr.test(node.name)


class AttrQuery(EntryQuery):
    pass


class AllAttrQuery(AttrQuery):
    def test(self, n):
        return all(self.expr.test(a) for a in n.attrs)


class AnyAttrQuery(AttrQuery):
    def test(self, n):
        return any(self.expr.test(a) for a in n.attrs)


def any_(*exprs):
    return AnyAttrQuery(Any(*[desugar_attr(e) for e in exprs]))


def all_(*exprs):
    return AllAttrQuery(All(*[desugar_attr(e) for e in exprs]))


def desugar_name(q):
    if q is None:
        return NameQuery(TRUE)
    if isinstance(q, NameQuery):
        return q
    if isinstance(q, Boolean):
        return NameQuery(q)
    if callable(q):
        return NameQuery(lift(q))
    return NameQuery(lift(partial(operator.eq, q)))


def desugar_attr(q):
    if isinstance(q, Boolean):
        return q
    if callable(q):
        return lift(q)
    return lift(partial(operator.eq, q))


def desugar_attrs(q):
    if not q:
        return
    if len(q) == 1:
        q = q[0]
        return q if isinstance(q, AttrQuery) else AnyAttrQuery(desugar_attr(q))
    else:
        attr_queries = [desugar_attr(a) for a in q[1:]]
        return AnyAttrQuery(Any(*attr_queries))


def desugar(q):
    if isinstance(q, tuple):
        q = list(q)
        name_query = desugar_name(q[0])
        attrs_query = desugar_attrs(q[1:])
        if attrs_query:
            return All(name_query, attrs_query)
        return name_query
    return desugar_name(q)


def flatten(nodes):
    def inner(n):
        res = [n]
        res.extend(chain.from_iterable(inner(c) for c in n.children))
        return res
    return list(chain.from_iterable(inner(n) for n in nodes))


def compile_queries(*queries):
    def match(qs, nodes):
        q = desugar(qs[0])
        res = [n for n in nodes if q.test(n)]
        qs = qs[1:]
        if qs:
            gc = list(chain.from_iterable(n.children for n in res))
            return match(qs, gc)
        return res

    def inner(nodes):
        return Result(children=match(queries, nodes))
    return inner


def find(query, nodes, roots=False):
    results = []
    for n in flatten(nodes):
        results.extend(query([n]))
    if not roots:
        return Result(children=results)

    seen = set()
    top = []
    for r in results:
        root = r.root
        if root not in seen:
            seen.add(root)
            top.append(root)
    return Result(children=top)


lt = lift2(operator.lt)
le = lift2(operator.le)
eq = lift2(operator.eq)
gt = lift2(operator.gt)
ge = lift2(operator.ge)

contains = lift2(operator.contains)
startswith = lift2(str.startswith)
endswith = lift2(str.endswith)

ieq = lift2(operator.eq, ignore_case=True)
icontains = lift2(operator.contains, ignore_case=True)
istartswith = lift2(str.startswith, ignore_case=True)
iendswith = lift2(str.endswith, ignore_case=True)



"""
iniparser parses ini files into a python dictionary of dictionaries. Duplicate
keys in a section convert the value to a list. Numbers are automatically
converted to python ints or floats. Line continuations based on hanging indents
are supported. Sections inherit keys from the `[DEFAULT]` section. All keys are
converted to lower case. Variable interpolation is not supported.
"""
import string
from parsr import (EOF, HangingString, InSet, LeftBracket, LineEnd, Many,
                   Number, OneLineComment, Opt, RightBracket, String,
                   WithIndent, WS, WSChar)


def loads(s):
    res = Top(s)[0]
    default = res.get("default")
    if default:
        default_keys = set(default)
        for header, items in res.items():
            if header != "default":
                missing = default_keys - set(items)
                for m in missing:
                    items[m] = default[m]

    return res


def load(f):
    return loads(f.read())


def to_dict(x):
    x = [i for i in x if i is not None]
    d = {}
    for k, v in x:
        k = k.lower()
        if k in d:
            if not isinstance(d[k], list):
                d[k] = [d[k]]
            d[k].append(v)
        else:
            d[k] = v
    return d


header_chars = (set(string.printable) - set(string.whitespace) - set("[]"))
sep_chars = set("=:")
key_chars = header_chars - sep_chars
value_chars = set(string.printable) - set("\n\r")

Num = Number & (WSChar | LineEnd)
LeftEnd = (WS + LeftBracket + WS)
RightEnd = (WS + RightBracket + WS)
Header = LeftEnd >> String(header_chars) << RightEnd
Key = WS >> String(key_chars) << WS
Sep = InSet(sep_chars)
Value = WS >> (Num | HangingString(value_chars))
KVPair = WithIndent(Key + Opt(Sep + Value, default=[None, None])).map(lambda a: (a[0], a[1][1]))
Comment = (WS >> (OneLineComment("#") | OneLineComment(";")).map(lambda x: None))
Line = Comment | KVPair
Section = Header + (Many(Line).map(to_dict))
Doc = Many(Comment | Section).map(to_dict)
Top = Doc + EOF



import string


# A way of describing arithmetic:

# number = Many(digits)
# factor = ("(" + expr + ")") | number
# term = factor + Many("*/" + factor)
# expr = term + Many("+-" + term)


def number(pos, data):
    # number = Many(digits).map(to_int)
    rest = []
    while data[pos] and data[pos] in string.digits:
        rest.append(data[pos])
        pos += 1
    return pos, int("".join(rest))


def factor(pos, data):
    # factor = ("(" + expr + ")") | number
    if data[pos] == "(":
        pos, res = expr(pos + 1, data)
        assert data[pos] == ")", "Unmatched paren."
        return pos + 1, res
    return number(pos, data)


def term(pos, data):
    # term = factor + Many("*/" + factor)
    pos, left = factor(pos, data)
    while data[pos] and data[pos] in "*/":
        op = data[pos]
        pos, right = factor(pos + 1, data)
        if op == "*":
            left *= right
        else:
            left /= right
    return pos, left


def expr(pos, data):
    # expr = term + Many("+-" + term)
    pos, left = term(pos, data)
    while data[pos] and data[pos] in "+-":
        op = data[pos]
        pos, right = term(pos + 1, data)
        if op == "+":
            left += right
        else:
            left -= right
    return pos, left


def evaluate(data):
    data = list(data)
    data.append(None)
    return expr(0, data)[1]



"""
The boolean module allows delayed evaluation of boolean expressions. You wrap
predicates in objects that have overloaded operators so they can be connected
symbolically to express "and", "or", and "not". This is useful if you want to
build up a complicated predicate and pass it to something else for evaluation.

```python
def is_even(n):
    return (n % 2) == 0

def is_positive(n):
    return n > 0

even_and_positive = lift(is_even) & lift(is_positive)

even_and_positive(6) == True
even_and_positive(-2) == False
even_and_positive(3) == False
```

You can also lift two parameter functions to which you want to partially
apply an argument. The arguments partially applied will be those *after*
the first argument. The first argument is the value the function should
evaluate when it's fully applied.

```python
import operator
lt = lift2(operator.lt)  # operator.lt is lt(a, b) == (a < b)
gt = lift2(operator.gt)  # operator.gt is gt(a, b) == (a > b)

gt_five = gt(5)  # creates a function of one argument that when called
                 # returns operator.gt(x, 5)

lt_ten = lt(10)  # creates a function of one argument that when called
                 # returns operator.lt(x, 5)

gt_five_and_lt_10 = gt(5) & lt(10)
```
"""


class Boolean:
    def __and__(self, other):
        return All(self, other)

    def __or__(self, other):
        return Any(self, other)

    def __invert__(self):
        return Not(self)

    def test(self, value):
        return True

    def __call__(self, value):
        return self.test(value)


class TRUE(Boolean):
    pass


class FALSE(Boolean):
    def test(self, value):
        return False


class Any(Boolean):
    def __init__(self, *exprs):
        self.exprs = list(exprs)

    def __or__(self, other):
        self.exprs.append(other)
        return self

    def test(self, value):
        return any(q.test(value) for q in self.exprs)


class All(Boolean):
    def __init__(self, *exprs):
        self.exprs = list(exprs)

    def __and__(self, other):
        self.exprs.append(other)
        return self

    def test(self, value):
        return all(q.test(value) for q in self.exprs)


class Not(Boolean):
    def __init__(self, query):
        self.query = query

    def test(self, value):
        return not self.query.test(value)


class Lift(Boolean):
    def __init__(self, func, *args):
        self.func = func
        self.args = args

    def test(self, value):
        try:
            return self.func(value, *self.args)
        except Exception:
            return False


class CaselessLift(Lift):
    def test(self, lhs):
        if isinstance(lhs, str):
            return super().test(lhs.lower())
        return super().test(lhs)


def lift(func, ignore_case=False):
    if ignore_case:
        return CaselessLift(func)
    return Lift(func)


def lift2(func, ignore_case=False):
    def inner(val):
        if ignore_case:
            return CaselessLift(func, val.lower())
        return Lift(func, val)
    return inner


Or = Any
And = All



import string


# A way of describing arithmetic:

# number = Many(digits)
# factor = ("(" + expr + ")") | number
# term = factor + Many("*/" + factor)
# expr = term + Many("+-" + term)


def oper(args):
    # externalize the operations to a function
    left, rest = args
    for op, right in rest:
        if op == "*":
            left *= right
        elif op == "/":
            left /= right
        elif op == "+":
            left += right
        else:
            left -= right
    return left


def number(pos, data):
    # number = Many(digits).map(to_int)

    # Many(string.digits)
    digits = many(inset(string.digits))
    p = _map(digits, lambda x: int("".join(x)))
    return p(pos, data)


def factor(pos, data):
    # factor = ("(" + expr + ")") | number

    # the lambda plucks out the value between ( and )
    subexpr = _map(seq(char("("), expr, char(")")), lambda x: x[1])
    p = choice(subexpr, number)
    return p(pos, data)


def term(pos, data):
    # term = (factor + Many("*/" + factor)).map(oper)
    p = _map(seq(factor, many(seq(inset("*/"), factor))), oper)
    return p(pos, data)


def expr(pos, data):
    # expr = (term + Many("+-" + term)).map(oper)
    p = _map(seq(term, many(seq(inset("+-"), term))), oper)
    return p(pos, data)


def evaluate(data):
    data = list(data)
    data.append(None)
    return expr(0, data)[1]


def char(c):
    # match a char
    def process(pos, data):
        if data[pos] == c:
            return pos + 1, c
        raise Exception()
    return process


def inset(s):
    # match anything in a set
    def process(pos, data):
        v = set(s)
        if data[pos] in v:
            return pos + 1, data[pos]
        raise Exception()
    return process


def many(p):
    # continue matching until failure
    def process(pos, data):
        rest = []
        while True:
            try:
                pos, r = p(pos, data)
                rest.append(r)
            except Exception:
                break
        return pos, rest
    return process


def seq(*args):
    # match several things in a row
    def process(pos, data):
        rest = []
        for p in args:
            pos, r = p(pos, data)
            rest.append(r)
        return pos, rest
    return process


def choice(*args):
    # match one of several alternatives
    def process(pos, data):
        e = None
        for p in args:
            try:
                return p(pos, data)
            except Exception as ex:
                e = ex
        raise e
    return process


def _map(p, func):
    # apply a function to the result of a parser's match
    def process(pos, data):
        pos, res = p(pos, data)
        return pos, func(res)
    return process



from setuptools import setup, find_packages

develop = set([
    'ipython',
    'pytest',
    'setuptools',
    'twine',
    'wheel',
])


if __name__ == "__main__":
    setup(
        name="parsr",
        version="0.2.4",
        description="Parsr is a simple parser combinator library in pure python.",
        long_description=open("README.md").read(),
        long_description_content_type='text/markdown',
        url="https://github.com/csams/parsr",
        packages=find_packages(),
        package_data={'': ['LICENSE']},
        license='Apache 2.0',
        extras_require={
            'develop': list(develop),
        },
        classifiers=[
            'Intended Audience :: Developers',
            'Natural Language :: English',
            'License :: OSI Approved :: Apache Software License',
            'Programming Language :: Python',
            'Programming Language :: Python :: 3.6',
            'Programming Language :: Python :: 3.7'
        ],
        include_package_data=True
    )



import string


# A way of describing arithmetic:

# number = Many(digits)
# factor = ("(" + expr + ")") | number
# term = factor + Many("*/" + factor)
# expr = term + Many("+-" + term)


class Parser:
    def __add__(self, other):
        return Seq(self, other)

    def __or__(self, other):
        return Choice(self, other)

    def map(self, func):
        return Map(self, func)

    def __call__(self, data):
        data = list(data)
        data.append(None)
        return self.process(0, data)[1]


# These classes correspond to functions of similar names in arith5.py. We use
# classes to take advantage of operator overloading.
class Char(Parser):
    # match a char
    def __init__(self, c):
        self.c = c

    def process(self, pos, data):
        if data[pos] == self.c:
            return pos + 1, self.c
        raise Exception()


class InSet(Parser):
    # match anything in a set
    def __init__(self, s):
        self.v = set(s)

    def process(self, pos, data):
        if data[pos] in self.v:
            return pos + 1, data[pos]
        raise Exception()


class Many(Parser):
    # continue matching until failure
    def __init__(self, p):
        self.p = p

    def process(self, pos, data):
        rest = []
        while True:
            try:
                pos, r = self.p.process(pos, data)
                rest.append(r)
            except Exception:
                break
        return pos, rest


class Seq(Parser):
    # match several things in a row
    def __init__(self, *args):
        self.args = list(args)

    def __add__(self, other):
        self.args.append(other)
        return self

    def process(self, pos, data):
        rest = []
        for p in self.args:
            pos, r = p.process(pos, data)
            rest.append(r)
        return pos, rest


class Choice(Parser):
    # match one of several alternatives
    def __init__(self, *args):
        self.args = list(args)

    def __or__(self, other):
        self.args.append(other)
        return self

    def process(self, pos, data):
        e = None
        for p in self.args:
            try:
                return p.process(pos, data)
            except Exception as ex:
                e = ex
        raise e


class Map(Parser):
    # apply a function to the result of a parser's match
    def __init__(self, p, func):
        self.p = p
        self.func = func

    def process(self, pos, data):
        pos, res = self.p.process(pos, data)
        return pos, self.func(res)


class Forward(Parser):
    # needed to "forward declare" some things since we're dealing with classes
    # and objects now instead of functions and can't just use them recursively.
    def __init__(self):
        self.p = None

    def __le__(self, other):
        self.p = other

    def process(self, pos, data):
        return self.p.process(pos, data)


def oper(args):
    left, rest = args
    for op, right in rest:
        if op == "*":
            left *= right
        elif op == "/":
            left /= right
        elif op == "+":
            left += right
        else:
            left -= right
    return left


expr = Forward()
number = Many(InSet(string.digits)).map(lambda x: int("".join(x)))
factor = (Char("(") + expr + Char(")")).map(lambda x: x[1]) | number
term = (factor + Many(InSet("*/") + factor)).map(oper)
expr <= (term + Many(InSet("+-") + term)).map(oper)


def evaluate(data):
    return expr(data)



import string


# A way of describing arithmetic:

# number = Many(digits)
# factor = ("(" + expr + ")") | number
# term = factor + Many("*/" + factor)
# expr = term + Many("+-" + term)


def oper(args):
    # externalize the operations to a function
    left, rest = args
    for op, right in rest:
        if op == "*":
            left *= right
        elif op == "/":
            left /= right
        elif op == "+":
            left += right
        else:
            left -= right
    return left


def number(pos, data):
    # number = Many(digits).map(to_int)

    # Many(string.digits)
    pos, res = many(pos, data, string.digits)
    # .map(lambda res: int("".join(res)))
    return pos, int("".join(res))


def factor(pos, data):
    # factor = ("(" + expr + ")") | number

    # (Char("(")
    if data[pos] == "(":
        # + expr
        pos, res = expr(pos + 1, data)
        # + Char(")"))
        assert data[pos] == ")", "Unmatched paren."
        return pos + 1, res
    # | number
    return number(pos, data)


def term(pos, data):
    # term = (factor + Many("*/" + factor)).map(oper)

    # factor
    pos, left = factor(pos, data)
    # + Many(Char("*/") + factor)
    pos, rest = many(pos, data, "*/", factor)
    # .map(oper)
    return pos, oper([left, rest])


def expr(pos, data):
    # expr = (term + Many("+-" + term)).map(oper)

    # term
    pos, left = term(pos, data)
    # + Many(Char("+-") + term)
    pos, rest = many(pos, data, "+-", term)
    # .map(oper)
    return pos, oper([left, rest])


def evaluate(data):
    data = list(data)
    data.append(None)
    return expr(0, data)[1]


def many(pos, data, cs, rhs=None):
    # ecapsulate the "build up a list" logic
    rest = []
    while data[pos] and data[pos] in cs:
        op = data[pos]
        if rhs:
            pos, right = rhs(pos + 1, data)
            rest.append([op, right])
        else:
            rest.append(op)
            pos += 1
    return pos, rest



import string
from bisect import bisect_left
from io import StringIO


class Node:
    def __init__(self):
        self.children = []

    def add_child(self, child):
        self.children.append(child)
        return self

    def set_children(self, children):
        self.children.clear()
        for c in children:
            self.add_child(c)
        return self

    def __repr__(self):
        return self.__class__.__name__


def text_format(tree):
    out = StringIO()
    tab = " " * 2
    seen = set()

    def inner(cur, prefix):
        print(prefix + str(cur), file=out)
        if cur in seen:
            return

        seen.add(cur)

        next_prefix = prefix + tab
        for c in cur.children:
            inner(c, next_prefix)

    inner(tree, "")
    out.seek(0)
    return out.read()


def render(tree):
    print(text_format(tree))


class Ctx:
    def __init__(self, lines):
        self.error_pos = -1
        self.error_msg = None
        self.indents = []
        self.tags = []
        self.lines = [i for i, x in enumerate(lines) if x == "\n"]

    def set(self, pos, msg):
        if pos >= self.error_pos:
            self.error_pos = pos
            self.error_msg = msg

    def line(self, pos):
        return bisect_left(self.lines, pos)

    def col(self, pos):
        p = self.line(pos)
        if p == 0:
            return pos
        return (pos - self.lines[p - 1] - 1)


class Parser(Node):
    def __init__(self):
        super().__init__()
        self.name = None

    def map(self, func):
        return Map(self, func)

    @staticmethod
    def accumulate(first, rest):
        results = [first] if first else []
        if rest:
            results.extend(rest)
        return results

    def sep_by(self, sep):
        return Lift(self.accumulate) * Opt(self) * Many(sep >> self)

    def until(self, pred):
        return Until(self, pred)

    def __or__(self, other):
        return Choice([self, other])

    def __and__(self, other):
        return FollowedBy(self, other)

    def __truediv__(self, other):
        return NotFollowedBy(self, other)

    def __add__(self, other):
        return Seq([self, other])

    def __lshift__(self, other):
        return KeepLeft(self, other)

    def __rshift__(self, other):
        return KeepRight(self, other)

    def __mod__(self, name):
        self.name = name
        return self

    def __call__(self, data):
        data = list(data)
        data.append(None)  # add a terminal so we don't overrun
        ctx = Ctx(data)
        ex = None
        try:
            _, ret = self.process(0, data, ctx)
            return ret
        except Exception:
            lineno = ctx.line(ctx.error_pos) + 1
            colno = ctx.col(ctx.error_pos) + 1
            ex = Exception(f"At line {lineno} column {colno}: {ctx.error_msg}")
        if ex:
            raise ex

    def __repr__(self):
        return self.name or f"{self.__class__.__name__}"


class Wrapper(Parser):
    def __init__(self, parser):
        super().__init__()
        self.add_child(parser)

    def process(self, pos, data, ctx):
        return self.children[0].process(pos, data, ctx)


class Mark:
    def __init__(self, lineno, col, value):
        self.lineno = lineno
        self.col = col
        self.value = value


class PosMarker(Wrapper):
    """
    Save the line number and column of a non-terminal or terminal by
    wrapping it in a PosMarker. The value of the parser that handled the input
    as well as the initial input position will be returned as a Mark instance.
    """
    def process(self, pos, data, ctx):
        lineno = ctx.line(pos) + 1
        col = ctx.col(pos) + 1
        pos, result = super().process(pos, data, ctx)
        return pos, Mark(lineno, col, result)


class Seq(Parser):
    def __init__(self, children):
        super().__init__()
        self.set_children(children)

    def __add__(self, other):
        return self.add_child(other)

    def process(self, pos, data, ctx):
        results = []
        for p in self.children:
            pos, res = p.process(pos, data, ctx)
            results.append(res)
        return pos, results


class Choice(Parser):
    def __init__(self, children):
        super().__init__()
        self.set_children(children)

    def __or__(self, other):
        return self.add_child(other)

    def process(self, pos, data, ctx):
        ex = None
        for c in self.children:
            try:
                return c.process(pos, data, ctx)
            except Exception as e:
                ex = e
        raise ex


class Many(Wrapper):
    def process(self, pos, data, ctx):
        results = []
        p = self.children[0]
        while True:
            try:
                pos, res = p.process(pos, data, ctx)
                results.append(res)
            except Exception:
                break
        return pos, results


class Many1(Many):
    def process(self, pos, data, ctx):
        pos, results = super().process(pos, data, ctx)
        if len(results) == 0:
            child = self.children[0]
            raise Exception(f"Expected at least one {child}.")
        return pos, results


class Until(Parser):
    def __init__(self, parser, predicate):
        super().__init__()
        self.set_children([parser, predicate])

    def process(self, pos, data, ctx):
        parser, pred = self.children
        results = []
        while True:
            try:
                pred.process(pos, data, ctx)
            except Exception:
                try:
                    pos, res = parser.process(pos, data, ctx)
                    results.append(res)
                except Exception:
                    break
            else:
                break
        return pos, results


class FollowedBy(Parser):
    def __init__(self, p, f):
        super().__init__()
        self.set_children([p, f])

    def process(self, pos, data, ctx):
        left, right = self.children
        new, res = left.process(pos, data, ctx)
        try:
            right.process(new, data, ctx)
        except Exception:
            raise
        else:
            return new, res


class NotFollowedBy(Parser):
    def __init__(self, p, f):
        super().__init__()
        self.set_children([p, f])

    def process(self, pos, data, ctx):
        left, right = self.children
        new, res = left.process(pos, data, ctx)
        try:
            right.process(new, data, ctx)
        except Exception:
            return new, res
        else:
            raise Exception(f"{right} can't follow {left}")


class KeepLeft(Parser):
    def __init__(self, left, right):
        super().__init__()
        self.set_children([left, right])

    def process(self, pos, data, ctx):
        left, right = self.children
        pos, res = left.process(pos, data, ctx)
        pos, _ = right.process(pos, data, ctx)
        return pos, res


class KeepRight(Parser):
    def __init__(self, left, right):
        super().__init__()
        self.set_children([left, right])

    def process(self, pos, data, ctx):
        left, right = self.children
        pos, _ = left.process(pos, data, ctx)
        pos, res = right.process(pos, data, ctx)
        return pos, res


class Opt(Parser):
    def __init__(self, p, default=None):
        super().__init__()
        self.add_child(p)
        self.default = default

    def process(self, pos, data, ctx):
        try:
            return self.children[0].process(pos, data, ctx)
        except Exception:
            return pos, self.default


class Map(Parser):
    def __init__(self, p, func):
        super().__init__()
        self.add_child(p)
        self.func = func

    def process(self, pos, data, ctx):
        pos, res = self.children[0].process(pos, data, ctx)
        return pos, self.func(res)

    def __repr__(self):
        return f"Map({self.func})"


class Lift(Parser):
    def __init__(self, func):
        super().__init__()
        self.func = func

    def __mul__(self, other):
        return self.add_child(other)

    def process(self, pos, data, ctx):
        ex = None
        results = []
        for c in self.children:
            pos, res = c.process(pos, data, ctx)
            results.append(res)
        try:
            return pos, self.func(*results)
        except Exception as e:
            ex = e
        if ex:
            ctx.set(pos, str(ex))
            raise ex


class Forward(Parser):
    def __init__(self):
        super().__init__()
        self.delegate = None

    def __le__(self, delegate):
        self.set_children([delegate])

    def process(self, pos, data, ctx):
        return self.children[0].process(pos, data, ctx)


class EOF(Parser):
    def process(self, pos, data, ctx):
        if data[pos] is None:
            return pos, None
        msg = "Expected end of input."
        ctx.set(pos, msg)
        raise Exception(msg)


class Char(Parser):
    def __init__(self, char):
        super().__init__()
        self.char = char

    def process(self, pos, data, ctx):
        if data[pos] == self.char:
            return (pos + 1, self.char)
        msg = f"Expected {self.char}."
        ctx.set(pos, msg)
        raise Exception(msg)

    def __repr__(self):
        return f"Char('{self.char}')"


class InSet(Parser):
    def __init__(self, s, name=None):
        super().__init__()
        self.values = set(s)
        self.name = name

    def process(self, pos, data, ctx):
        c = data[pos]
        if c in self.values:
            return (pos + 1, c)
        msg = f"Expected {self}."
        ctx.set(pos, msg)
        raise Exception(msg)


class Literal(Parser):
    _NULL = object()

    def __init__(self, chars, value=_NULL, ignore_case=False):
        super().__init__()
        self.chars = chars if not ignore_case else chars.lower()
        self.value = value
        self.ignore_case = ignore_case

    def process(self, pos, data, ctx):
        old = pos
        if not self.ignore_case:
            for c in self.chars:
                if data[pos] == c:
                    pos += 1
                else:
                    msg = f"Expected {self.chars}."
                    ctx.set(old, msg)
                    raise Exception(msg)
            return pos, (self.chars if self.value is self._NULL else self.value)
        else:
            result = []
            for c in self.chars:
                if data[pos].lower() == c:
                    result.append(data[pos])
                    pos += 1
                else:
                    msg = f"Expected case insensitive {self.chars}."
                    ctx.set(old, msg)
                    raise Exception(msg)
            return pos, ("".join(result) if self.value is self._NULL else self.value)


class String(Parser):
    def __init__(self, chars, echars=None):
        super().__init__()
        self.chars = set(chars)
        self.echars = set(echars) if echars else set()

    def process(self, pos, data, ctx):
        results = []
        p = data[pos]
        old = pos
        while p in self.chars or p == "\\":
            if p == "\\" and data[pos + 1] in self.echars:
                results.append(data[pos + 1])
                pos += 2
            elif p in self.chars:
                results.append(p)
                pos += 1
            else:
                break
            p = data[pos]
        if not results:
            msg = f"Expected one of {self.chars}."
            ctx.set(old, msg)
            raise Exception(msg)
        return pos, "".join(results)


class EnclosedComment(Parser):
    def __init__(self, s, e):
        super().__init__()
        Start = Literal(s)
        End = Literal(e)
        p = Start >> AnyChar.until(End).map(lambda x: "".join(x)) << End
        self.add_child(p)

    def process(self, pos, data, ctx):
        return self.children[0].process(pos, data, ctx)


class OneLineComment(Parser):
    def __init__(self, s):
        super().__init__()
        p = Literal(s) >> Opt(String(set(string.printable) - set("\r\n")), "")
        self.add_child(p)

    def process(self, pos, data, ctx):
        return self.children[0].process(pos, data, ctx)


class WithIndent(Wrapper):
    def process(self, pos, data, ctx):
        new, _ = WS.process(pos, data, ctx)
        try:
            ctx.indents.append(ctx.col(new))
            return self.children[0].process(new, data, ctx)
        finally:
            ctx.indents.pop()


class HangingString(Parser):
    def __init__(self, chars):
        super().__init__()
        self.add_child(String(chars) << (EOL | EOF))

    def process(self, pos, data, ctx):
        old = pos
        results = []
        while True:
            try:
                if ctx.col(pos) > ctx.indents[-1]:
                    pos, res = self.children[0].process(pos, data, ctx)
                    results.append(res)
                else:
                    pos = old
                    break
                old = pos
                pos, _ = WS.process(pos, data, ctx)
            except Exception:
                break
        ret = " ".join(results)
        return pos, ret


class StartTagName(Wrapper):
    def process(self, pos, data, ctx):
        pos, res = self.children[0].process(pos, data, ctx)
        ctx.tags.append(res)
        return pos, res


class EndTagName(Wrapper):
    def process(self, pos, data, ctx):
        pos, res = self.children[0].process(pos, data, ctx)
        expect = ctx.tags.pop()
        if res != expect:
            msg = f"Expected {expect}. Got {res}."
            ctx.set(pos, msg)
            raise Exception(msg)
        return pos, res


def make_number(sign, int_part, frac_part):
    tmp = sign + int_part + ("".join(frac_part) if frac_part else "")
    return float(tmp) if "." in tmp else int(tmp)


def skip_none(x):
    return [i for i in x if i is not None]


EOF = EOF()
EOL = InSet("\n\r") % "EOL"
LineEnd = Wrapper(EOL | EOF)
EQ = Char("=")
LT = Char("<")
GT = Char(">")
FS = Char("/")
LeftCurly = Char("{")
RightCurly = Char("}")
LeftBracket = Char("[")
RightBracket = Char("]")
LeftParen = Char("(")
RightParen = Char(")")
Colon = Char(":")
SemiColon = Char(";")
Comma = Char(",")
AnyChar = InSet(string.printable) % "Any Char"
NonZeroDigit = InSet(set(string.digits) - set("0"))
Digit = InSet(string.digits) % "Digit"
Digits = String(string.digits) % "Digits"
Letter = InSet(string.ascii_letters)
Letters = String(string.ascii_letters)
WSChar = InSet(set(string.whitespace) - set("\n\r")) % "Whitespace w/o EOL"
WS = Many(InSet(string.whitespace)) % "Whitespace"
Number = (Lift(make_number) * Opt(Char("-"), "") * Digits * Opt(Char(".") + Digits)) % "Number"
SingleQuotedString = Char("'") >> String(set(string.printable) - set("'"), "'") << Char("'")
DoubleQuotedString = Char('"') >> String(set(string.printable) - set('"'), '"') << Char('"')
QuotedString = Wrapper(DoubleQuotedString | SingleQuotedString) % "Quoted String"
